from abc import ABC, abstractmethod
from typing import Any


class BaseSerializer(ABC):
    """
    Responsible for message serialization and data validation.

    Different serializers may be compatible, but better
    have serializers of the same type on both sides.
    """

    __slots__ = ()

    @abstractmethod
    def dump(self, data: Any) -> bytes:
        raise NotImplementedError

    @abstractmethod
    def load(self, raw_data: bytes) -> Any:
        raise NotImplementedError

    # distinct methods if serialization on client side differs

    def client_dump(self, data: Any) -> bytes:
        return self.dump(data)

    def client_load(self, raw_data: bytes) -> Any:
        return self.load(raw_data)



from .base import BaseSerializer
from .json import JsonSerializer
from .adaptix import AdaptixSerializer


default_serializer = JsonSerializer()


__all__ = (
    'BaseSerializer',
    'JsonSerializer',
    'AdaptixSerializer'
)



from typing import Any

from communica.utils import json_dumpb, json_loadb

from .base import BaseSerializer


class JsonSerializer(BaseSerializer):
    def dump(self, data: Any) -> bytes:
        return json_dumpb(data)

    def load(self, raw_data: bytes) -> Any:
        return json_loadb(raw_data)



import random
import asyncio
import logging
from abc import abstractmethod
from enum import Enum
from uuid import uuid1, uuid4
from typing import Any, Generic, TypeVar, TypedDict, cast
from inspect import iscoroutinefunction
from traceback import format_exc
from dataclasses import dataclass

from communica.utils import (
    TaskSet,
    HasLoopMixin,
    BackoffDelayer,
    iscallable,
    fmt_task_name,
)
from communica.exceptions import ReqError, RespError, UnknownError, SerializerError
from communica.serializers import BaseSerializer, default_serializer
from communica.entities.base import (
    BaseClient,
    BaseServer,
    SyncHandlerType,
    AsyncHandlerType,
)
from communica.connectors.base import (
    HandshakeOk,
    HandshakeGen,
    BaseConnector,
    BaseConnection,
)


__all__ = (
    'SimpleClient',
    'SimpleServer'
)


FlowT = TypeVar('FlowT', bound='ReqRepMessageFlow')


logger = logging.getLogger('communica.entities.simple')


class RequestType(int, Enum):
    REQ_REP = 1
    REQ_THROW = 2

    RESP_OK = 11

    RESP_ERR_UNKNOWN = 30

    RESP_ERR_REQUESTER = 41
    RESP_ERR_DATA_LOAD = 42

    RESP_ERR_RESPONDER = 51


@dataclass
class ServerHandshakeOk(HandshakeOk):
    client_id: str


class Metadata(TypedDict):
    type: RequestType
    id: str


class RequestHandler:
    __slots__ = ('_repr', 'is_async', 'endpoint', 'running_tasks')

    is_async: bool
    endpoint: 'SyncHandlerType | AsyncHandlerType'
    running_tasks: TaskSet

    def __repr__(self) -> str:
        try:
            return self._repr
        except AttributeError:
            name = getattr(self.endpoint, '__qualname__',
                           getattr(self.endpoint, '__name__', 'UNKNOWN'))
            htype = 'async' if self.is_async else 'sync'
            self._repr = f'<RequestHandler {htype} endpoint={name}>'
            return self._repr

    def __init__(self, endpoint: 'SyncHandlerType | AsyncHandlerType') -> None:
        if not iscallable(endpoint):
            raise TypeError('Request handler must be function, '
                            'coroutine function or method')
        self.is_async = iscoroutinefunction(endpoint)
        self.endpoint = endpoint
        self.running_tasks = TaskSet()


class ReqRepMessageFlow(HasLoopMixin):
    __slots__ = (
        '_connection', '_response_waiters'
    )

    _response_waiters: 'dict[str, asyncio.Future]'

    @property
    def connection(self):
        return self._connection

    def __init__(self):
        self._response_waiters = {}

    def update_connection(self, connection: BaseConnection):
        if hasattr(self, '_connection'):
            self._connection.update(connection)
        else:
            self._connection = connection
        return self._connection

    def _create_response_waiter(self):
        request_id = uuid1().hex

        fut = self._get_loop().create_future()
        self._response_waiters[request_id] = fut
        fut.add_done_callback(lambda _: self._response_waiters.pop(request_id, None))

        return request_id, fut

    def _handle_response(
            self,
            serializer: BaseSerializer,
            metadata: Metadata,
            raw_data: bytes
    ):
        fut = self._response_waiters.pop(metadata['id'], None)
        if fut is None or fut.done():
            logger.warning('Dropped response for unknown or expired request')
            return

        req_type = metadata['type']

        if req_type == RequestType.RESP_OK:
            try:
                fut.set_result(serializer.client_load(raw_data))
            except Exception as e:
                # if serializer can't load response, considering this
                # requester's error, cause responder can't do
                # anything about it after sending response
                fut.set_exception(e)
            return

        data = default_serializer.client_load(raw_data)
        if req_type == RequestType.RESP_ERR_REQUESTER:
            fut.set_exception(ReqError.from_dict(data))
        elif req_type == RequestType.RESP_ERR_RESPONDER:
            fut.set_exception(RespError.from_dict(data))
        elif req_type == RequestType.RESP_ERR_UNKNOWN:
            fut.set_exception(UnknownError.from_dict(data))
        elif req_type == RequestType.RESP_ERR_DATA_LOAD:
            fut.set_exception(SerializerError.from_dict(data))
        else:
            logger.critical(f"{metadata = }, wtf")
            fut.set_exception(
                UnknownError(f'Got unknown response type: {metadata["type"]}')
            )

    async def _handle_request(
            self,
            handler: RequestHandler,
            serializer: BaseSerializer,
            metadata: Metadata,
            raw_data: Any
    ) -> 'tuple[dict, Any]':
        resp_meta = metadata.copy()

        try:
            data = serializer.load(raw_data)
        except Exception as e:
            resp_meta['type'] = RequestType.RESP_ERR_DATA_LOAD
            resp_data = SerializerError(repr(e)).to_dict()
            return resp_meta, resp_data  # type: ignore

        try:
            if handler.is_async:
                resp_data = await handler.endpoint(data)
            else:
                resp_data = handler.endpoint(data)
            resp_meta['type'] = RequestType.RESP_OK

        except ReqError as e:
            resp_data = e.to_dict()
            resp_meta['type'] = RequestType.RESP_ERR_REQUESTER

        except RespError as e:
            resp_data = e.to_dict()
            resp_meta['type'] = RequestType.RESP_ERR_RESPONDER

        except Exception as e:
            logger.exception('Unexpected exception in %r', handler)
            resp_data = UnknownError(repr(e)).to_dict()
            resp_meta['type'] = RequestType.RESP_ERR_UNKNOWN

        return resp_meta, resp_data  # type: ignore

    async def _send_response(
            self,
            req_meta,
            resp_meta,
            resp_data,
            serializer: BaseSerializer
    ):
        if req_meta['type'] == RequestType.REQ_REP:
            if resp_meta['type'] > RequestType.RESP_OK:
                serializer = default_serializer

            try:
                raw_data = serializer.dump(resp_data)
            except Exception:
                logger.error('Can\'t serialize response:\n%s', format_exc())

                raw_data = default_serializer.dump(
                    RespError('Response serialize error').to_dict()
                )
                resp_meta['type'] = RequestType.RESP_ERR_RESPONDER

            await self._connection.send(resp_meta, raw_data)

        elif resp_meta['type'] > RequestType.RESP_ERR_UNKNOWN:
            logger.warning('Error occured while handling request, but '
                           'requester don\'t know about this:\n' + resp_data['msg'])

    @abstractmethod
    def dispatch(self, metadata: Any, raw_data: bytes): ...


class SimpleMessageFlow(ReqRepMessageFlow):
    __slots__ = ('handler', 'serializer')

    def __init__(
            self,
            handler: 'SyncHandlerType | AsyncHandlerType',
            serializer: BaseSerializer
    ):
        super().__init__()
        self.handler = RequestHandler(handler)
        self.serializer = serializer

    def dispatch(self, metadata: Metadata, raw_data: bytes):
        if metadata['type'] < RequestType.RESP_OK:
            self.handler.running_tasks.create_task_with_exc_log(
                self.handle_request(metadata, raw_data),
                name=fmt_task_name('simple-request-handler')
            )
        else:
            self._handle_response(self.serializer, metadata, raw_data)

    async def handle_request(self, req_meta: Metadata, raw_data: bytes):
        resp_meta, resp_data = await self._handle_request(
            self.handler, self.serializer, req_meta, raw_data
        )
        await self._send_response(
            req_meta, resp_meta, resp_data, self.serializer
        )

    async def request(self, data: Any) -> bytes:
        request_id, fut = self._create_response_waiter()

        await self._connection.send(
            Metadata(id=request_id, type=RequestType.REQ_REP),
            self.serializer.client_dump(data)
        )

        return await fut

    async def throw(self, data: Any) -> None:
        await self._connection.send(
            Metadata(id='', type=RequestType.REQ_THROW),
            self.serializer.client_dump(data)
        )


class ReqRepClient(BaseClient, Generic[FlowT]):
    __slots__ = ('_connected_event', '_run_task', '_client_id', '_flow')

    _flow: FlowT
    _run_task: 'asyncio.Task | None'

    @property
    def client_id(self) -> str:
        return self._client_id

    @property
    def is_running(self):
        return not (self._run_task is None or self._run_task.done())

    @property
    def connected_event(self) -> asyncio.Event:
        try:
            return self._connected_event
        except AttributeError:
            self._connected_event = asyncio.Event()
            return self._connected_event

    async def init(self, timeout: 'int | None' = None):
        if not self._run_task or self._run_task.done():
            self._run_task = self._get_loop().create_task(
                self._connection_keeper(),
                name=fmt_task_name('client-connection-keeper')
            )

        try:
            await asyncio.wait_for(self.connected_event.wait(), timeout)
        except asyncio.TimeoutError:
            self._run_task.cancel()
            raise
        return self

    async def close(self) -> None:
        if self._run_task is not None:
            await self._flow._connection.close()
            self._run_task.cancel()

    async def _connection_keeper(self):
        delayer = BackoffDelayer(0.1, 5, 2, 0.5)
        while True:
            try:
                new_conn = await self.connector.client_connect(self._handshaker)
            except Exception as e:
                logger.warning('%r: Connect failed: %r', self.connector, e)
                await delayer.wait()
                continue

            connection = self._flow.update_connection(new_conn)
            self.connected_event.set()
            delayer.reset()

            logger.info('Connected to %s', self.connector.repr_address())
            try:
                await connection.run_until_fail(self._flow.dispatch)
            except Exception as e:
                logger.exception('Unhandled exception '
                                 'in connection runner: %r', e)
            self.connected_event.clear()
            logger.info('Disconnected from server on %s',
                        self.connector.repr_address())
            await asyncio.sleep(1)

    async def _handshaker(self, connection: BaseConnection) -> HandshakeGen:
        client_hello = {
            'client_id': self._client_id
        }

        server_hello = (yield client_hello)  # noqa

        yield HandshakeOk()


class SimpleClient(ReqRepClient):
    """
    Pair to SimpleServer.

    Has only one optional request handler.
    """

    __slots__ = ('serializer',)

    serializer: BaseSerializer
    _flow: SimpleMessageFlow

    def __init__(
            self,
            connector: BaseConnector,
            handler: 'SyncHandlerType | AsyncHandlerType | None' = None,
            serializer: 'BaseSerializer | None' = None,
            client_id: 'str | None' = None,
    ) -> None:
        super().__init__(connector)

        if serializer is None:
            serializer = default_serializer
        if handler is None:
            handler = self._not_defined_handler

        self._flow = SimpleMessageFlow(handler, serializer)

        self._client_id = client_id or uuid4().hex
        self._run_task = None

    def _not_defined_handler(self, data: Any):
        raise RespError('Client side did not define handler for server requests')

    async def request(self, data: Any) -> Any:
        """Send request, wait response."""
        return await self._flow.request(data)

    async def throw(self, data: Any) -> None:
        """Send request without waiting response."""
        return await self._flow.throw(data)


class ReqRepServer(BaseServer, Generic[FlowT]):
    __slots__ = ('_server', '_known_clients', '_client_conn_runners',
                 '_client_connected')

    _server: asyncio.AbstractServer
    _client_conn_runners: 'dict[str, asyncio.Task]'

    # future waiter waits specific client
    _known_clients: 'dict[str, FlowT | asyncio.Future[FlowT]]'
    # event waiter waits any client
    _client_connected: asyncio.Event

    async def init(self):
        if not hasattr(self, '_server') or not self._server.is_serving():
            self._server = await self.connector.server_start(
                self._handshaker, self._on_client_connect)
        return self

    async def close(self) -> None:
        if not hasattr(self, '_server'):
            return

        self._server.close()
        await self._server.wait_closed()

        for client_id, flow in self._known_clients.items():
            if isinstance(flow, ReqRepMessageFlow):
                self._cancel_handler_tasks(flow)
                await flow._connection.close()
            if (conn_task := self._client_conn_runners.get(client_id)):
                conn_task.cancel()

    async def _handshaker(self, connection: BaseConnection) -> HandshakeGen:
        server_hello = {
            'hello': 'hello'
        }

        client_hello = (yield server_hello)

        client_id = client_hello['client_id']
        yield ServerHandshakeOk(client_id=client_id)

    # TODO: clear _known_clients
    async def _get_client_flow(
            self,
            client_id: 'str | None'
    ) -> FlowT:
        if client_id is None:
            connected_clients = self._get_connected_clients()
            if not connected_clients:
                await self._client_connected.wait()
            return random.choice(connected_clients)

        flow = self._known_clients.get(client_id)
        if not isinstance(flow, ReqRepMessageFlow):
            if flow is None:
                self._known_clients[client_id] = self._get_loop().create_future()
            flow = await asyncio.shield(self._known_clients[client_id])  # pyright: ignore
        return flow

    def _on_client_connect(self, connection: BaseConnection):
        loop = self._get_loop()
        handshake_result = cast(ServerHandshakeOk, connection.get_handshake_result())
        client_id = handshake_result.client_id

        flow = self._known_clients.get(client_id)
        if not isinstance(flow, ReqRepMessageFlow):
            new_flow = self._create_new_flow()
            if isinstance(flow, asyncio.Future):
                flow.set_result(new_flow)
            self._known_clients[client_id] = new_flow
            flow = new_flow

        connection = flow.update_connection(connection)
        self._client_connected.set()
        self._client_connected.clear()

        task = loop.create_task(
            connection.run_until_fail(flow.dispatch),
            name=fmt_task_name('connection-runner')
        )
        task.add_done_callback(self._on_conn_fail)
        self._client_conn_runners[client_id] = task
        logger.info('Client with id %r connected to server on %s',
                    client_id, self.connector.repr_address())

    def _get_connected_clients(self):
        return [
            flow for flow in self._known_clients.values()
                if not isinstance(flow, asyncio.Future) and flow.connection.is_alive
        ]

    def _on_conn_fail(self, task: asyncio.Task):
        if task.cancelled():
            return
        if (exc := task.exception()):
            logger.warning(f'Client read failed: {exc!r}')

    @abstractmethod
    def _cancel_handler_tasks(self, flow: FlowT):
        raise NotImplementedError

    @abstractmethod
    def _create_new_flow(self) -> FlowT:
        raise NotImplementedError


class SimpleServer(ReqRepServer[SimpleMessageFlow]):
    """
    Pair to SimpleClient.

    Has only one request handler.
    """

    __slots__ = ('_handler', '_serializer')

    _handler: 'SyncHandlerType | AsyncHandlerType'
    _serializer: BaseSerializer

    def __init__(
            self,
            connector: BaseConnector,
            handler: 'SyncHandlerType | AsyncHandlerType',
            serializer: 'BaseSerializer | None' = None,
    ) -> None:
        if serializer is None:
            serializer = default_serializer

        self._handler = handler
        self.connector = connector
        self._serializer = serializer
        self._known_clients = {}
        self._client_connected = asyncio.Event()
        self._client_conn_runners = {}

    def _cancel_handler_tasks(self, flow: SimpleMessageFlow):
        flow.handler.running_tasks.cancel()

    def _create_new_flow(self) -> SimpleMessageFlow:
        return SimpleMessageFlow(self._handler, self._serializer)

    async def request(self, data: Any, client_id: 'str | None' = None) -> Any:
        """
        Send request, wait for response.

        Args:
            client_id: If omitted or None, random connected client will be chosen.
        """
        flow = await self._get_client_flow(client_id)
        return await flow.request(data)

    async def throw(self, data: Any, client_id: 'str | None' = None) -> None:
        """
        Send request without waiting for response.

        Args:
            client_id: If omitted or None, random connected client will be chosen.
        """
        flow = await self._get_client_flow(client_id)
        return await flow.throw(data)



import asyncio
import logging
import weakref
from abc import ABC, abstractmethod
from enum import Enum
from uuid import uuid4
from typing import Any
from collections import deque

from typing_extensions import Self

from communica.utils import (
    NULL_CHAR,
    HasLoopMixin,
    json_dumpb,
    json_loadb,
    fmt_task_name,
    read_accessor,
    exc_log_callback,
)
from communica.exceptions import FeatureNotAvailable
from communica.connectors.base import (
    Handshaker,
    BaseConnector,
    HandshakeFail,
    BaseConnection,
    ClientConnectedCB,
    RequestReceivedCB,
)
from communica.connectors.rabbitmq.pool import (
    ChannelPool,
    ChannelRoute,
    PooledChannel,
    ChannelExpired,
)


try:
    import aiormq
    _HAVE_AIORMQ = True
except ModuleNotFoundError:
    _HAVE_AIORMQ = False
else:
    from yarl import URL


DEFAULT_POOL = ChannelPool()
DEFAULT_EXCHANGE = 'communica'
HANDSHAKE_TIMEOUT = 30

logger = logging.getLogger('communica.connectors.rabbitmq')


class _MessageType(str, Enum):
    CONNECT_REQUEST = 'communica_client_connect'
    CONNECT_RESPONSE = 'communica_client_connect_ok'

    HS_NEXT = 'communica_handshake_next'
    HS_FAIL = 'communica_handshake_fail'
    HS_DONE = 'communica_handshake_done'

    CLOSE = 'close'
    MESSAGE = 'message'
    LISTENING = 'listening'


class MessageWaiter(HasLoopMixin):
    __slots__ = ('_waiter', '_messages', '_timeout')

    @classmethod
    async def new(
            cls,
            chan: PooledChannel,
            queue: str,
            *,
            no_ack: bool,
            timeout: 'int | None' = None
    ):
        inst = cls()
        inst._waiter = None
        inst._timeout = timeout
        inst._messages = deque()

        await chan.consume(queue, inst._callback, no_ack=no_ack)

        return inst

    async def _callback(self, message: 'aiormq.abc.DeliveredMessage'):
        if not self._waiter or self._waiter.done():
            self._messages.append(message)
        else:
            self._waiter.set_result(message)
            self._waiter = None

    def set_timeout_duration(self, timeout: int):
        if self._waiter and not self._waiter.done():
            raise RuntimeError('Waiter not completed')
        self._timeout = timeout

    def _set_timeout(self, fut: asyncio.Future):
        if not fut.done():
            fut.set_exception(asyncio.TimeoutError)

    def interrupt(self):
        if self._waiter and not self._waiter.done():
            self._waiter.set_exception(asyncio.TimeoutError)

    async def wait(self) -> 'aiormq.abc.DeliveredMessage':
        if self._messages:
            return self._messages.popleft()

        if self._waiter and not self._waiter.done():
            raise RuntimeError(f'Duplicate .wait() call on {self!r}')

        self._waiter = self._get_loop().create_future()

        if self._timeout:
            self._get_loop().call_later(
                self._timeout, self._set_timeout, self._waiter)

        return await self._waiter


class ConnectionCheckPolicy(HasLoopMixin, ABC):
    period: float
    _handle: 'asyncio.TimerHandle | None'

    def message_sent(self):
        pass

    def message_received(self):
        pass

    @abstractmethod
    def _trigger(self):
        raise NotImplementedError

    @abstractmethod
    def cancel(self):
        raise NotImplementedError

    @abstractmethod
    def replace_conn(self, conn: 'RmqConnection') -> Self:
        raise NotImplementedError

    def _set_handle(self, period: 'float | None' = None):
        if period is None:
            self._last_message = self._get_loop().time()
            period = self.period
        self._handle = self._get_loop().call_later(period, self._trigger)


class ServerCheckPolicy(ConnectionCheckPolicy):
    def __init__(self, conn: 'RmqConnection', period: float) -> None:
        self.period = period
        self._waiter = self._get_loop().create_future()

        self._send_task = self._get_loop().create_task(
            self._sender(),
            name=fmt_task_name('rmq-server-listen-notifier')
        )
        self._send_task.add_done_callback(exc_log_callback)
        self._conn = weakref.ref(conn, self._conn_died)

        self._set_handle()

    def replace_conn(self, conn: 'RmqConnection'):
        conn._check_policy.cancel()
        self._conn = weakref.ref(conn, self._conn_died)
        return self

    def cancel(self):
        if self._handle is not None:
            self._handle.cancel()
        self._conn_died(None)
        self._send_task.cancel()

    def _conn_died(self, _):
        if not self._waiter.done():
            self._waiter.set_result(False)

    def _trigger(self):
        if self._waiter.done():
            return
        if self._send_task.done():
            return

        time_diff = self._get_loop().time() - self._last_message
        if time_diff < self.period:
            self._set_handle(self.period - time_diff)
            return

        self._handle = None
        self._waiter.set_result(True)
        self._waiter = self._get_loop().create_future()

    async def _sender(self):
        while (await self._waiter):
            if (conn := self._conn()) is None:
                return
            await conn._send(_MessageType.LISTENING, b'')
            del(conn)  # deleting reference
            if self._handle is None:
                self._set_handle()

    def message_sent(self):
        if self._handle is None:
            self._set_handle()
        else:
            self._last_message = self._get_loop().time()


class ClientCheckPolicy(ConnectionCheckPolicy):
    def __init__(self, conn: 'RmqConnection', period: float) -> None:
        self.period = period
        self._set_handle()

        self._conn = weakref.ref(conn)
        self._close_task = None

    def replace_conn(self, conn: 'RmqConnection'):
        conn._check_policy.cancel()
        self._conn = weakref.ref(conn)
        return self

    def cancel(self):
        if self._handle is not None:
            self._handle.cancel()

    def _trigger(self):
        if (conn := self._conn()) is None:
            return

        time_diff = self._get_loop().time() - self._last_message
        if time_diff < self.period:
            self._set_handle(self.period - time_diff)
            return

        logger.info('Closing %r cause no messages received in %f seconds',
                    conn, self.period)
        self._close_task = self._get_loop().create_task(
            conn.close(),
            name=fmt_task_name('rmq-client-check-closer')
        )

    def message_received(self):
        if self._close_task is None:
            self._last_message = self._get_loop().time()


class RmqConnection(BaseConnection):
    __slots__ = ('_connector', '_chan', '_ready', '_closing',
                 '_recv_queue', '_send_route', '_check_policy',
                 '_connect_id', '__weakref__')

    _chan: PooledChannel
    _ready: asyncio.Event
    _closing: 'asyncio.Future | None'
    _connector: 'RmqConnector'

    @property
    def is_alive(self):
        return self._ready.is_set()

    @classmethod
    async def _do_handshake_and_create_connection(
            cls,
            connector: 'RmqConnector',
            handshaker: Handshaker,
            resp_waiter: MessageWaiter,
            send_route: ChannelRoute
    ) -> Self:
        inst = cls()

        async def send_message(data: bytes):
            await send_route.publish(_MessageType.HS_NEXT, body=data)

        async def recv_message():
            try:
                message = await resp_waiter.wait()
            except asyncio.TimeoutError:
                raise HandshakeFail('Response timeout exceeded '
                                    'when connection is established') from None

            if message.header.properties.message_type == _MessageType.HS_FAIL:
                raise HandshakeFail.loadb(message.body)
            elif message.header.properties.message_type != _MessageType.HS_NEXT:
                raise ValueError('Unknown message in handshake queue')
            return message.body

        try:
            await inst._run_handshaker(handshaker, send_message, recv_message)
        except HandshakeFail as fail:
            await send_route.publish(_MessageType.HS_FAIL, fail.dumpb())
            raise

        inst._ready = asyncio.Event()
        inst._ready.set()
        inst._closing = None
        inst._connector = connector

        return inst

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        if self._ready:
            state = f'on channel {self._chan.number}'
        else:
            state = 'not ready'
        return f'<{cls_name} {state}>'

    def _set_run_data(
            self,
            chan: PooledChannel,
            policy: ConnectionCheckPolicy,
            recv_queue: str,
            send_queue: str,
            connect_id: bytes
    ):
        self._chan = chan
        self._connect_id = connect_id
        self._check_policy = policy
        self._recv_queue = recv_queue
        self._send_route = ChannelRoute(
            chan,
            self._connector.exchange_name,
            send_queue
        )

    def update(self, connection: Self) -> None:
        if self._connector._client_id != connection._connector._client_id:
            raise ValueError('Got connection with different client_id')
        if connection._chan.expired:
            raise ValueError('Got connection with expired channel')

        self._ready.set()
        self._closing = None
        self._set_run_data(
            connection._chan,
            connection._check_policy.replace_conn(self),
            connection._recv_queue,
            connection._send_route.routing_key,
            connection._connect_id,
        )

    async def send(self, metadata: Any, raw_data: bytes):
        body = json_dumpb(metadata) + NULL_CHAR + raw_data
        try:
            await self._send(_MessageType.MESSAGE, body)
        except ChannelExpired:
            self._ready.clear()
            await self._chan.release()
            await self._send(_MessageType.MESSAGE, body)  # retry once after channel opened
        except aiormq.exceptions.AMQPError:
            self._ready.clear()
            raise
        self._check_policy.message_sent()

    async def _send(self, msg_type: _MessageType, body: bytes):
        await self._ready.wait()
        logger.debug('%r sent %r to %r',
                     self, msg_type, self._send_route.routing_key)

        # cancelling of aiormq RPC task will send Channel.Close to server,
        # which greatly increases chances of connection exception,
        # which in turn will break everything on same URL in same pool
        await asyncio.shield(
            self._send_route.publish(msg_type, body=body)
        )

    async def run_until_fail(
            self,
            request_received_cb: RequestReceivedCB
    ) -> None:
        try:
            await self._run_until_fail(request_received_cb)
        except asyncio.TimeoutError:
            return
        finally:
            self._ready.clear()
            await self._chan.release()

    async def _run_until_fail(self, request_received_cb):
        waiter = await MessageWaiter.new(
            self._chan,
            self._recv_queue,
            no_ack=False
        )

        self._chan.on_release(lambda _: waiter.interrupt())
        self._chan.inner.closing.add_done_callback(lambda _: waiter.interrupt())

        while self._ready.is_set():
            message = await waiter.wait()
            logger.debug(
                '%r received %r from %r',
                self, message.header.properties.message_type, self._recv_queue
            )

            self._check_policy.message_received()

            if message.header.properties.message_type == _MessageType.MESSAGE:
                null_pos = message.body.find(NULL_CHAR)
                if null_pos == -1:
                    await self._chan.ack_message(message)
                    await self._close(True)
                    raise ValueError('Got message without metadata separator')

                metadata = json_loadb(message.body[:null_pos])
                request_received_cb(metadata, message.body[null_pos+1:])
                await self._chan.ack_message(message)

            elif message.header.properties.message_type == _MessageType.LISTENING:
                await self._chan.ack_message(message)
                continue

            elif message.header.properties.message_type == _MessageType.CLOSE:
                await self._chan.ack_message(message)

                if message.body == self._connect_id:
                    await self._close(False)
                    return

            else:
                await self._chan.ack_message(message)
                await self._close(True)
                message_type = message.header.properties.message_type
                raise ValueError('Unknown message in message queue, '
                                f'got type {message_type!r}')

    async def close(self) -> None:
        logger.debug('Closing %r', self)
        await self._close(True)

    async def _close(self, send_close_request: bool):
        if self._closing:
            return await self._closing

        if not self._ready.is_set():
            return

        self._ready.clear()
        self._check_policy.cancel()

        if self._chan.expired:
            return

        self._closing = asyncio.get_running_loop().create_future()
        try:
            if send_close_request:
                await self._send_route.publish(
                    _MessageType.CLOSE,
                    body=self._connect_id
                )
            await self._chan.release()
        except (
            ChannelExpired,
            asyncio.CancelledError,
            aiormq.exceptions.ChannelClosed,
            aiormq.exceptions.ConnectionClosed
        ):
            return
        finally:
            self._closing.set_result(None)


class RmqServer(asyncio.AbstractServer):
    @property
    def exchange(self):
        return self._connector._exchange

    def __init__(
            self,
            connector: 'RmqConnector',
            handshaker: Handshaker,
            client_connected_cb: ClientConnectedCB
    ) -> None:
        self._closing = None
        self._connector = connector
        self._handshaker = handshaker
        self._client_connected_cb = client_connected_cb

    def is_serving(self):
        """Returns True if the server accepts connections."""
        try:
            return not self._chan.expired
        except AttributeError:
            return False

    def close(self):
        if not hasattr(self, '_chan'):
            return

        async def close_channel():
            if not self._chan.expired:
                await self._chan.release()
            if self._closing is not asyncio.current_task():
                raise RuntimeError('Server closing task changed')
            self._closing = None

        if not self._closing or self._closing.done():
            self._closing = asyncio.create_task(
                close_channel(),
                name=fmt_task_name('rmq-server-closer')
            )

    async def wait_closed(self):
        if not self._closing or self._closing.done():
            raise RuntimeError('wait_closed() should be '
                               'called right after close() method')
        await self._closing

    async def _on_connect(self, message: 'aiormq.abc.DeliveredMessage'):
        if message.header.properties.message_type != _MessageType.CONNECT_REQUEST:
            await self._chan.ack_message(message)
            logger.error('Unknown message type in connect request queue (%r in %r)',
                         message.header.properties.message_type, self._queue_name)
            return

        if message.header.properties.reply_to is None:
            await self._chan.ack_message(message)
            logger.error('Reply queue is not specified in the connect request')
            return

        client_id = json_loadb(message.body)
        hs_to_cli_route = ChannelRoute(
            self._chan,
            self.exchange,
            message.header.properties.reply_to
        )

        logger.debug('got connect request from client_id=%s', client_id)

        hs_to_srv_queue = await self._chan.queue_declare_and_bind(
            self.exchange,
            exclusive=True
        )

        await hs_to_cli_route.publish(
            _MessageType.CONNECT_RESPONSE,
            body=b'',
            properties_kwargs={
                'reply_to': hs_to_srv_queue
            }
        )

        await self._chan.ack_message(message)

        run_chan = await self._connector.acquire_channel()
        resp_waiter = await MessageWaiter.new(
            run_chan,
            hs_to_srv_queue,
            no_ack=True,
            timeout=HANDSHAKE_TIMEOUT
        )
        hs_completed = False
        try:
            conn = await RmqConnection._do_handshake_and_create_connection(
                self._connector,
                self._handshaker,
                resp_waiter,
                hs_to_cli_route
            )
            hs_completed = True
        except HandshakeFail:
            # if fail happens on server's side, client will get and log it
            # so log on server only when debugging
            logger.debug('Handshake with client_id=%s failed',
                         client_id, exc_info=True)
            await run_chan.release()
            return
        finally:
            if not hs_completed:
                await run_chan.release()
            if not self._chan.expired:
                await self._chan.inner.queue_delete(hs_to_srv_queue)
        await run_chan.cancel_consume()

        to_cli_queue, to_srv_queue = \
            self._connector._get_transport_queue_names(client_id)

        await run_chan.queue_declare_and_bind(
            self.exchange,
            name=to_cli_queue,
            durable=True
        )
        await run_chan.queue_declare_and_bind(
            self.exchange,
            name=to_srv_queue,
            durable=True
        )

        connect_id = uuid4().bytes
        await hs_to_cli_route.publish(_MessageType.HS_DONE, connect_id)
        policy = ServerCheckPolicy(
            conn,
            self._connector.CONNECTION_CHECK_PERIOD
        )
        conn._set_run_data(
            run_chan,
            policy,
            to_srv_queue,
            to_cli_queue,
            connect_id
        )

        logger.info('Handshake with client_id=%s succedeed, connect_id=%s',
                     client_id, connect_id)
        self._client_connected_cb(conn)

    async def _restart_serving(self):
        logger.warning(
            'aiormq channel under %r was unexpectedly closed! '
            'Trying to restart server on a new channel',
            self
        )
        await self._chan.release()
        await self.start_serving()

    def _make_aiormq_channel_close_cb(self):
        last_alive_channel = self._chan

        def callback(_):
            if self._chan is not last_alive_channel:
                return  # server was restarted after pool closed
            if self._chan.explicitly_released:
                return  # server was closed explicitly
            asyncio.create_task(
                self._restart_serving(),
                name=fmt_task_name('rmq-server-restart')
            )

        return callback

    async def start_serving(self):
        if self.is_serving():
            return

        self._chan = await self._connector.acquire_channel()
        self._queue_name = self._connector._get_connect_queue_name()

        await self._chan.inner.basic_qos(prefetch_count=1)
        await self._chan.queue_declare_and_bind(
            self.exchange,
            self._queue_name,
            durable=True
        )
        await self._chan.consume(self._queue_name, self._on_connect)

        self._chan.inner.closing.add_done_callback(
            self._make_aiormq_channel_close_cb()
        )

    def get_loop(self):
        raise NotImplementedError

    async def serve_forever(self):
        raise NotImplementedError


class RmqConnector(BaseConnector):
    """
    Uses RabbitMQ for communication. Theoretically can be used
    with any AMQP-compatible server, but this wasn't tested.

    WARNING:
        For this connector to work properly, only one client
        with same address and client_id must be connected at same time.

        For those, who connect, client_id must be set explicitly
        and persist between service restarts.

        Failure to comply with these rules
        leads to unclosed queues growth and message loss.
    """

    _TYPE = 'RABBITMQ'

    CONNECTION_CHECK_PERIOD = 20
    """every X seconds server sends message to client,
    if client did not got any message in (this period * 1.5),
    it closes connection"""

    __slots__ = ('_url', '_address', '_client_id',
                 '_exchange', '_exchange_declared', '_pool')

    url: read_accessor[str] = read_accessor('_url')
    address: read_accessor[str] = read_accessor('_address')
    client_id: read_accessor[str] = read_accessor('_client_id')
    exchange_name: read_accessor[str] = read_accessor('_exchange')

    def __init__(
            self,
            url: str,
            address: str,
            client_id: 'str | None' = None,
            exchange_name: 'str | None' = None,
            channel_pool: 'ChannelPool | None' = None
    ) -> None:
        """
        Max summary length of address and client_id is 224 characters

        Args:
            url (str): used to create connection with RabbitMQ server.
              Format specified at https://www.rabbitmq.com/uri-spec.html.
            address (str): unique identifier for client-server pair.
              If more than one server with same address bound to same exchange,
              behaviour undefined.
            client_id (str): unique identifier of connecting process.
              Must be set for clients.
        """
        if not _HAVE_AIORMQ:
            raise FeatureNotAvailable(
                'RmqConnector requires aiormq and yarl libraries. '
                'Install communica with [rabbitmq] extra.'
            )

        if client_id is not None and len(address + client_id) > 224:
            raise ValueError('Max address + client_id length is 224 characters')

        self._url = URL(url)
        self._client_id = client_id
        self._address = address
        self._exchange_declared = False

        self._pool = channel_pool or DEFAULT_POOL
        self._exchange = exchange_name or DEFAULT_EXCHANGE

    def _get_connect_queue_name(self):
        return self._fmt_queue_name('connect', 'server')

    def _get_transport_queue_names(self, client_id: str):
        return (
            self._fmt_queue_name('toClient', client_id),
            self._fmt_queue_name('toServer', client_id)
        )

    def _fmt_queue_name(self, *parts):
        return f'communica.{self._address}.' + '.'.join(parts)

    def repr_address(self) -> str:
        if self._url.password is not None:
            return self._url.with_password("[PASSWORD]").human_repr()
        return self._url.human_repr()

    async def server_start(
            self,
            handshaker: Handshaker,
            client_connected_cb: ClientConnectedCB,
    ) -> asyncio.AbstractServer:
        chan = await self._pool.acquire_chan(self._url)

        await chan.inner.exchange_declare(
            self._exchange,
            exchange_type="direct",
            durable=True,
        )
        await chan.release()

        server = RmqServer(self, handshaker, client_connected_cb)
        await server.start_serving()
        return server

    async def _client_connect(
            self,
            handshaker: Handshaker,
            chan: PooledChannel
    ):
        await chan.inner.basic_qos(prefetch_count=1)

        hs_to_cli_queue = await chan.queue_declare_and_bind(
            self.exchange_name,
            exclusive=True
        )

        await chan.publish(
            msg_type=_MessageType.CONNECT_REQUEST,
            body=json_dumpb(self._client_id),
            exchange=self.exchange_name,
            routing_key=self._get_connect_queue_name(),
            properties_kwargs={
                'reply_to': hs_to_cli_queue
            }
        )

        resp_waiter = await MessageWaiter.new(
            chan,
            hs_to_cli_queue,
            no_ack=True
            # there is no timeout needed for first response,
            # server will reply as soon as it ready
        )

        message = await resp_waiter.wait()
        if message.header.properties.message_type != _MessageType.CONNECT_RESPONSE:
            raise ValueError('Unknown message in connect response queue')
        if message.header.properties.reply_to is None:
            raise ValueError('Server queue not specified in connect response')
        hs_to_srv_queue = message.header.properties.reply_to

        resp_waiter.set_timeout_duration(HANDSHAKE_TIMEOUT)

        try:
            conn = await RmqConnection._do_handshake_and_create_connection(
                self,
                handshaker,
                resp_waiter,
                ChannelRoute(chan, self.exchange_name, hs_to_srv_queue)
            )
        except Exception:
            if not chan.expired:
                await chan.inner.queue_delete(hs_to_cli_queue)
            raise

        message = await resp_waiter.wait()
        await chan.inner.queue_delete(hs_to_cli_queue)
        await chan.cancel_consume()

        if message.header.properties.message_type != _MessageType.HS_DONE:
            message_type = message.header.properties.message_type
            raise ValueError(f'Got message with "{message_type}" type '
                              'instead of handshake confirmation')
        connect_id = message.body

        to_cli_queue, to_srv_queue = \
            self._get_transport_queue_names(self._client_id)  # pyright: ignore[reportArgumentType]
        policy = ClientCheckPolicy(
            conn,
            self.CONNECTION_CHECK_PERIOD * 1.5
        )
        conn._set_run_data(
            chan,
            policy,
            to_cli_queue,
            to_srv_queue,
            connect_id
        )

        return conn

    async def client_connect(self, handshaker: Handshaker) -> BaseConnection:
        if self._client_id is None:
            raise TypeError('Cannot connect to server. For those, who connect, '
                            'client_id parameter must be set. '
                            'Check RmqConnector docs for details.')

        chan = await self.acquire_channel()
        try:
            if not self._exchange_declared:
                await chan.inner.exchange_declare(
                    self._exchange,
                    exchange_type="direct",
                    durable=True,
                )
                self._exchange_declared = True

            return await self._client_connect(handshaker, chan)
        except Exception:
            if not chan.expired:
                await chan.release()
            raise

    async def cleanup(self):
        """
        Drop all pending messages and
        delete queues with connector's client_id

        If client_id is None, server connect queue is deleted.
        """
        if self._client_id is None:
            queue_names = [self._get_connect_queue_name()]
        else:
            queue_names = self._get_transport_queue_names(self._client_id)

        for queue in queue_names:
            chan = await self.acquire_channel()
            try:
                await chan.inner.queue_purge(queue)
                await chan.inner.queue_delete(queue)
            except aiormq.exceptions.ChannelNotFoundEntity:
                # channel will be closed by server if queue is not found
                if not chan.expired:
                    await asyncio.wait([chan.inner.closing])
            await chan.release()

    def acquire_channel(self):
        return self._pool.acquire_chan(self._url)

    @classmethod
    def _load_pickled(cls, kwargs: dict):
        return cls(**kwargs)

    def __reduce__(self):
        return (
            RmqConnector._load_pickled,
            (
                {
                    'url': self._url,
                    'address': self._address,
                    'client_id': self._client_id,
                    'exchange_name': self._exchange,
                    'channel_pool': None,
                },
            )
        )



class CommunicaError(Exception):
    """Base exception for all library errors."""


class FeatureNotAvailable(CommunicaError):
    """
    Some feature is not available in current environment.

    Meant for lack of platform specific functionality
    or not installed libraries.
    """


class SerializerError(CommunicaError):
    """Error on loading or dumping data"""

    def __init__(self, message: str) -> None:
        super().__init__(message)

    @property
    def message(self):
        return self.args[0]

    def to_dict(self):
        return {
            'message': self.message
        }

    @classmethod
    def from_dict(cls, data: dict):
        return cls(data['message'])


class _RequestsErr(CommunicaError):
    def __init__(self, message: str, code: 'int | None' = None) -> None:
        super().__init__(message, code)

    @property
    def message(self):
        return self.args[0]

    @property
    def code(self):
        return self.args[1]

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.to_dict()!r})'

    def to_dict(self):
        return {
            'code': self.code,
            'message': self.message
        }

    @classmethod
    def from_dict(cls, data: dict):
        return cls(data['message'], data['code'])


class ReqError(_RequestsErr):
    """
    Errors, caused by requesting side
    (Server can make request to Client too)

    Can be raised by user code in request handler, in this case
    requesting side get same exception with specified
    message and code properties.
    """


class RespError(_RequestsErr):
    """
    Errors, caused by responding side.

    Can be raised by user code in request handler, in this case
    requesting side get same exception with specified
    message and code properties.
    """


class UnknownError(_RequestsErr):
    """Unhandled exception on responding side."""

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.message})'

    @property
    def code(self) -> None:
        return None



try:
    import adaptix

    _HAVE_ADAPTIX = True
except ModuleNotFoundError:
    _HAVE_ADAPTIX = False

from typing import Any, Type, Generic, TypeVar

from communica.utils import json_dumpb, json_loadb
from communica.exceptions import FeatureNotAvailable
from communica.serializers import BaseSerializer


TReq = TypeVar('TReq')
TResp = TypeVar('TResp')


class AdaptixSerializer(BaseSerializer, Generic[TReq, TResp]):
    __slots__ = ('_retort', '_req_model', '_resp_model')

    FROM_HANDLER = object()

    # XXX: один из вариантов
    # FROM_HANDLER = специальное значение
    # если регистратор ручки видит это значение вместо сериалайзера,
    # он берет ручку и через какой-нибудь метод собирает сериалайзер
    # по сигнатуре ручки

    def __init__(
            self,
            request_model: Type[TReq],
            response_model: 'Type[TResp] | None' = None,
            retort: 'adaptix.Retort | None' = None
    ) -> None:
        if not _HAVE_ADAPTIX:
            raise FeatureNotAvailable(
                'AdaptixSerializer requires adaptix library. '
                'Install communica with [adaptix] extra.'
            )

        self._retort = retort or adaptix.Retort()
        self._req_model = request_model
        self._resp_model = response_model or Any

    def load(self, raw_data: bytes) -> TReq:
        return self._retort.load(
            json_loadb(raw_data),
            self._req_model
        )

    def dump(self, data: TResp) -> bytes:
        return json_dumpb(
            self._retort.dump(data, self._resp_model)
        )

    def client_load(self, raw_data: bytes) -> 'TResp | Any':
        return self._retort.load(
            json_loadb(raw_data),
            self._resp_model
        )

    def client_dump(self, data: TReq) -> bytes:
        return json_dumpb(
            self._retort.dump(data, self._req_model)
        )



# type: ignore
import asyncio
import weakref


try:
    import nats
    _HAVE_NATS = True
except ModuleNotFoundError:
    _HAVE_NATS = False
else:
    import nats.errors
    import nats.aio.msg
    import nats.js.client
    import nats.aio.client
    import nats.aio.subscription
    from yarl import URL

from abc import ABC, abstractmethod
from enum import Enum
from typing import Any
from datetime import datetime, timedelta
from collections import deque

from typing_extensions import Self

from communica.utils import (
    NULL_CHAR,
    HasLoopMixin,
    json_dumpb,
    json_loadb,
    exc_log_callback,
)
from communica.connectors.base import (
    Handshaker,
    BaseConnector,
    HandshakeFail,
    BaseConnection,
    ClientConnectedCB,
    RequestReceivedCB,
)


_DEFAULT_PREFIX = 'communica'


class _MessageType(str, Enum):
    CONNECT_REQUEST = 'communica_client_connect'
    CONNECT_RESPONSE = 'communica_client_connect_ok'

    HS_NEXT = 'communica_handshake_next'
    HS_FAIL = 'communica_handshake_fail'
    HS_DONE = 'communica_handshake_done'

    CLOSE = 'close'
    MESSAGE = 'message'
    LISTENING = 'listening'


_clients: '''dict[
    tuple[asyncio.AbstractEventLoop, tuple[str, ...]],
    nats.aio.client.Client | asyncio.Future[nats.aio.client.Client]
]''' = {}


def _not_ready_exceptions():
    return (  # XXX: по семантике первое от второго и третьего слишком отличается
        nats.errors.ConnectionClosedError,  # pyright: ignore[reportUnboundVariable]
        nats.errors.ConnectionDrainingError,  # pyright: ignore[reportUnboundVariable]
        nats.errors.ConnectionReconnectingError  # pyright: ignore[reportUnboundVariable]
    )


async def _publish(
        client: 'nats.aio.client.Client | nats.js.client.JetStreamContext',
        msg_type: str,
        subject: str,
        body: 'bytes' = b'',
        reply: 'str | None' = None
):
    print('PUB', msg_type, body)
    headers = {'message_type': msg_type}
    if reply:
        headers['reply'] = reply
    return await client.publish(
        subject=subject,
        payload=body,
        headers=headers
    )


class Message:
    __slots__ = ('msg')

    def __init__(self, msg: 'nats.aio.msg.Msg') -> None:
        self.msg = msg

    @property
    def body(self) -> bytes:
        return self.msg.data

    @property
    def type(self) -> str:
        return self.msg.headers['message_type']  # type: ignore

    @property
    def reply(self) -> 'str | None':
        return self.msg.headers.get('reply')  # type: ignore



class MessageWaiter(HasLoopMixin):
    __slots__ = ('_waiter', '_messages', '_timeout', '_subscription')

    _waiter: 'asyncio.Future[Message] | None'
    _timeout: 'float | None'
    _messages: 'deque[Message]'
    _subscription: 'nats.aio.subscription.Subscription'

    @classmethod
    async def new(
            cls,
            client: 'nats.aio.client.Client | nats.js.client.JetStreamContext',
            subject: str,
            *,
            no_ack: bool = False, # XXX
            timeout: 'float | None' = None
    ):
        inst = cls()
        inst._waiter = None
        inst._timeout = timeout
        inst._messages = deque()

        inst._subscription = \
            await client.subscribe(subject, cb=inst._callback)

        return inst

    async def _callback(self, message: 'nats.aio.msg.Msg'):
        print('GOT', message.header, message.data)

        if not self._waiter or self._waiter.done():
            self._messages.append(Message(message))
        else:
            self._waiter.set_result(Message(message))
            self._waiter = None

    def set_timeout_duration(self, timeout: int):
        if self._waiter and not self._waiter.done():
            raise RuntimeError('Waiter not completed')
        self._timeout = timeout

    def _set_timeout(self, fut: asyncio.Future):
        if not fut.done():
            fut.set_exception(asyncio.TimeoutError)

    def interrupt(self):
        if self._waiter:
            self._waiter.set_exception(asyncio.TimeoutError)

    async def unsubscribe(self):
        await self._subscription.unsubscribe()

    async def wait(self) -> 'Message':
        if self._messages:
            return self._messages.popleft()

        if self._waiter and not self._waiter.done():
            raise RuntimeError(f'Duplicate .wait() call on {self!r}')

        self._waiter = self._get_loop().create_future()

        if self._timeout:
            self._get_loop().call_later(
                self._timeout, self._set_timeout, self._waiter)

        return await self._waiter


class ConnectionCheckPolicy(HasLoopMixin, ABC):
    period: float
    _handle: 'asyncio.TimerHandle | None'

    def message_sent(self):
        pass

    def message_received(self):
        pass

    @abstractmethod
    def _trigger(self):
        raise NotImplementedError

    @abstractmethod
    def _cancel(self):
        raise NotImplementedError

    @abstractmethod
    def replace_conn(self, conn: 'NATSConnection') -> Self:
        raise NotImplementedError

    def _set_handle(self, period: 'float | None' = None):
        if period is None:
            self._last_message = self._get_loop().time()
            period = self.period
        self._handle = self._get_loop().call_later(period, self._trigger)


class ServerCheckPolicy(ConnectionCheckPolicy):
    def __init__(self, conn: 'NATSConnection', period: float) -> None:
        self._get_loop()

        self.period = period
        self._waiter = self._get_loop().create_future()

        self._set_handle()

        self._send_task = self._get_loop().create_task(self._sender())
        self._send_task.add_done_callback(exc_log_callback)
        self._conn = weakref.ref(conn, self._conn_died)

    def replace_conn(self, conn: 'NATSConnection'):
        conn._check_policy._cancel()
        self._conn = weakref.ref(conn, self._conn_died)
        return self

    def _cancel(self):
        if self._handle is not None:
            self._handle.cancel()
        self._conn_died(None)

    def _conn_died(self, _):
        if not self._waiter.done():
            self._waiter.set_result(False)

    def _trigger(self):
        if self._waiter.done():
            return

        time_diff = self._get_loop().time() - self._last_message
        if time_diff < self.period:
            self._set_handle(self.period - time_diff)
            return

        self._handle = None
        self._waiter.set_result(True)
        self._waiter = self._get_loop().create_future()

    async def _sender(self):
        while (await self._waiter):
            if (conn := self._conn()) is None:
                return
            await conn._send(_MessageType.LISTENING, b'')
            del(conn)  # deleting reference
            if self._handle is None:
                self._set_handle()

    def message_sent(self):
        if self._handle is None:
            self._set_handle()
        else:
            self._last_message = self._get_loop().time()


class ClientCheckPolicy(ConnectionCheckPolicy):
    def __init__(self, conn: 'NATSConnection', period: float) -> None:
        self._get_loop()

        self.period = period
        self._set_handle()

        self._conn = weakref.ref(conn)
        self._close_task = None

    def replace_conn(self, conn: 'NATSConnection'):
        conn._check_policy._cancel()
        self._conn = weakref.ref(conn)
        return self

    def _cancel(self):
        if self._handle is not None:
            self._handle.cancel()

    def _trigger(self):
        if (conn := self._conn()) is None or not conn.is_alive:
            return

        time_diff = self._get_loop().time() - self._last_message
        if time_diff < self.period:
            self._set_handle(self.period - time_diff)
            return

        self._close_task = self._get_loop().create_task(conn.close())

    def message_received(self):
        if self._close_task is None:
            self._last_message = self._get_loop().time()


class NATSConnection(BaseConnection):
    __slots__ = ('_connector', '_js_context', '_ready', '_closing',
                 '_recv_subject', '_send_subject', '_check_policy',
                 '_message_waiter', '__weakref__')

    _ready: asyncio.Event
    _closing: 'asyncio.Future | None'
    _connector: 'NATSConnector'
    _js_context: 'nats.js.client.JetStreamContext'
    _message_waiter: 'MessageWaiter | None'

    @property
    def is_alive(self):
        return self._ready.is_set()

    @classmethod
    async def _do_handshake_and_create_connection(
            cls,
            js_context: 'nats.js.client.JetStreamContext',
            connector: 'NATSConnector',
            handshaker: Handshaker,
            resp_waiter: MessageWaiter,
            send_subject: str
    ) -> Self:
        inst = cls()

        async def send_message(data: bytes):
            await _publish(
                js_context,
                msg_type=_MessageType.HS_NEXT,
                body=data,
                subject=send_subject
            )

        async def recv_message():
            message = await resp_waiter.wait()
            if message.type == _MessageType.HS_FAIL:
                raise HandshakeFail.loadb(message.body)
            elif message.type != _MessageType.HS_NEXT:
                raise ValueError('Unknown message on handshake subject')
            return message.body

        try:
            await inst._run_handshaker(handshaker, send_message, recv_message)
        except HandshakeFail as fail:
            await _publish(
                js_context,
                msg_type=_MessageType.HS_FAIL,
                body=fail.dumpb(),
                subject=send_subject
            )
            raise

        inst._ready = asyncio.Event()
        inst._ready.set()
        inst._closing = None
        inst._connector = connector
        inst._js_context = js_context
        inst._message_waiter = None

        return inst

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        if self._ready:
            state = f'on {self._js_context}'
        else:
            state = 'not ready'
        return f'<{cls_name} {state}>'

    def _set_run_data(
            self,
            policy: ConnectionCheckPolicy,
            recv_subject: str,
            send_subject: str
    ):
        self._check_policy = policy
        self._recv_subject, self._send_subject = recv_subject, send_subject

    def update(self, connection: Self) -> None:
        if self._connector._connect_id != connection._connector._connect_id:
            raise ValueError('Got connection with different connect_id')
        # if connection._js_context.:
            raise ValueError('Got connection with closed client')

        self._ready.set()
        self._closing = None
        self._js_context = connection._js_context
        self._check_policy = connection._check_policy.replace_conn(self)

    async def send(self, metadata: Any, raw_data: bytes):
        body = json_dumpb(metadata) + NULL_CHAR + raw_data
        try:
            await self._send(_MessageType.MESSAGE, body)
        except _not_ready_exceptions():
            self._ready.clear()
            await self._send(_MessageType.MESSAGE, body)  # retry once after channel opened
        # except aiormq.exceptions.AMQPError:
        #     self._ready.clear()
        #     raise
        self._check_policy.message_sent()

    async def _send(self, msg_type: _MessageType, body: bytes):
        await self._ready.wait()
        await _publish(
            self._js_context,
            msg_type=msg_type,
            body=body,
            subject=self._send_subject
        )

    async def run_until_fail(
            self,
            request_received_cb: RequestReceivedCB
    ) -> None:
        self._message_waiter = await MessageWaiter.new(
            self._js_context,
            self._recv_subject,
            no_ack=False
        )
        try:
            await self._run_until_fail(request_received_cb)
        except asyncio.TimeoutError:
            return
        finally:
            self._message_waiter = None
            self._ready.clear()

    async def _run_until_fail(self, request_received_cb):
        while self._message_waiter:
            message = await self._message_waiter.wait()

            self._check_policy.message_received()

            if message.type == _MessageType.MESSAGE:
                null_pos = message.body.find(NULL_CHAR)
                if null_pos == -1:
                    # await ack_message(message)
                    await self._close(True)
                    raise ValueError('Got message without metadata separator')

                metadata = json_loadb(message.body[:null_pos])
                request_received_cb(metadata, message.body[null_pos+1:])
                # await ack_message(message)

            elif message.type == _MessageType.LISTENING:
                # await ack_message(message)
                continue

            elif message.type == _MessageType.CLOSE:
                # await ack_message(message)
                await self._close(False)
                return

            else:
                # await ack_message(message)
                await self._close(True)
                raise ValueError('Unknown message on message subject, '
                                f'got type {message.type!r}')

    async def close(self) -> None:
        await self._close(True)

    async def _close(self, send_close_request: bool):
        if self._closing:
            return await self._closing

        if not self._ready.is_set():
            return

        self._ready.clear()

        if self._message_waiter:
            self._message_waiter.interrupt()
            self._message_waiter = None

        if self._js_context.is_closed:
            return

        self._closing = asyncio.get_running_loop().create_future()
        try:
            if send_close_request:
                await _publish(
                    self._js_context,
                    msg_type=_MessageType.CLOSE,
                    subject=self._send_subject
                )
        except _not_ready_exceptions() + (asyncio.CancelledError,):
            return
        finally:
            await self._connector._check_connection_use()
            self._closing.set_result(None)


class NATSServer(asyncio.AbstractServer):
    @property
    def exchange(self):
        return self._connector._exchange

    def __init__(
            self,
            connector: 'NATSConnector',
            nats_client: 'nats.aio.client.Client',
            handshaker: Handshaker,
            client_connected_cb: ClientConnectedCB
    ) -> None:
        self._closing = None
        self._connector = connector
        self._handshaker = handshaker
        self._nats_client = nats_client
        self._client_connected_cb = client_connected_cb

    def is_serving(self):
        """Return True if the server is accepting connections."""
        return not self._nats_client.is_closed

    def close(self):
        async def close_channel():
            # ну типа надо придумать
            return

        if not self._closing or self._closing.done():
            self._closing = asyncio.create_task(close_channel())

    async def wait_closed(self):
        if not self._closing or self._closing.done():
            raise RuntimeError('wait_closed() should be '
                               'called right after close() method')
        await self._closing

    async def _on_connect(self, raw_message: 'nats.aio.msg.Msg'):
        message = Message(raw_message)

        if message.type != _MessageType.CONNECT_REQUEST:
            # await self._connector._ack_message(self._chan, message)
            raise ValueError('Unknown message on connect request subject')

        if message.reply is None:
            # await self._connector._ack_message(self._chan, message)
            raise ValueError('Reply subject not specified in connect request')

        hs_to_cli_subject = message.reply
        connect_id = json_loadb(message.body)

        hs_to_srv_subject = self._nats_client.new_inbox()

        await _publish(
            self._nats_client,
            msg_type=_MessageType.CONNECT_RESPONSE,
            subject=hs_to_cli_subject,
            reply=hs_to_srv_subject
        )

        # await self._connector._ack_message(self._chan, message)

        resp_waiter = await MessageWaiter.new(
            self._nats_client,
            hs_to_srv_subject,
            no_ack=True,
            timeout=10
        )

        try:
            conn = await NATSConnection._do_handshake_and_create_connection(
                self._nats_client,
                self._connector,
                self._handshaker,
                resp_waiter,
                hs_to_cli_subject
            )
        except HandshakeFail:
            return
        finally:
            if not self._nats_client.is_closed:
                await resp_waiter.unsubscribe()

        to_cli_subject, to_srv_subject = \
            self._connector._get_transport_subject_names(connect_id)

        await _publish(
            self._nats_client,
            msg_type=_MessageType.HS_DONE,
            subject=hs_to_cli_subject
        )

        policy = \
            ServerCheckPolicy(conn, self._connector.CONNECTION_CHECK_PERIOD)
        conn._set_run_data(policy, to_srv_subject, to_cli_subject)

        self._client_connected_cb(conn)

    async def start_serving(self):
        subject_name = self._connector._get_connect_subject_name()
        await self._nats_client.subscribe(subject_name, cb=self._on_connect)

    def get_loop(self):
        raise NotImplementedError

    async def serve_forever(self):
        raise NotImplementedError


class NATSConnector(BaseConnector):
    """
    Uses RabbitMQ for communication. Theoretically can be used
    with any AMQP-compatible server, but this wasn't tested.

    WARNING:
        For this connector to work properly, only one client
        with same address and connect_id must be connected at same time.

        For those, who connect, connect_id must be set explicitly
        and persist between service restarts.

        Failure to comply with these rules
        leads to unclosed subjects growth and message loss.
    """

    _TYPE = 'RABBITMQ'

    CONNECTION_CHECK_PERIOD = 20
    """every X seconds server sends message to client,
    if client did not got any message in (this period * 1.5),
    it closes connection"""

    __slots__ = ('_urls', '_address', '_connect_id',
                 '_exchange', '_exchange_declared')

    @property
    def exchange(self):
        return self._exchange

    def __init__(
            self,
            urls: 'str | list[str]',
            address: str,
            connect_id: 'str | None' = None,
    ) -> None:
        """
        Max summary length of address and connect_id is 224 characters

        Args:
            urls (str): used to create connection with NATS servers.
              Format specified at https://www.rabbitmq.com/uri-spec.html.
            address (str): unique identifier for client-server pair.
              If more than one server with same address bound to same exchange,
              behaviour undefined.
            connect_id (str): unique identifier of connecting process.
              Must be set for clients.
        """
        if not _HAVE_NATS:
            raise ImportError('NATSConnector requires nats library. '
                              'Install communica with [nats] extra.')

        if connect_id is not None and len(address + connect_id) > 224:
            raise ValueError('Max address + connect_id length is 224 characters')

        if isinstance(urls, str):
            urls = [urls]
        self._urls = tuple(urls)
        self._address = address
        self._connect_id = connect_id

    def _get_connect_subject_name(self):
        return self._fmt_subject_name('connect', 'server')

    def _get_transport_subject_names(self, connect_id: str):
        return (
            self._fmt_subject_name('toClient', connect_id),
            self._fmt_subject_name('toServer', connect_id)
        )

    def _fmt_subject_name(self, *parts):
        return f'communica.{self._address}.' + '.'.join(parts)

    def repr_address(self) -> str:
        urls: 'list[URL]' = []
        for raw_url in self._urls:
            url = URL(raw_url)
            if url.password is not None:
                url = url.with_password("[PASSWORD]")
            urls.append(url)
        return ','.join([url.human_repr() for url in urls])

    async def server_start(
            self,
            handshaker: Handshaker,
            client_connected_cb: ClientConnectedCB,
    ) -> asyncio.AbstractServer:
        nats_client = await self._get_client()

        server = NATSServer(self, nats_client, handshaker, client_connected_cb)
        await server.start_serving()
        return server

    async def _client_connect(
            self,
            handshaker: Handshaker,
            nats_client: 'nats.aio.client.Client'
    ):
        hs_to_cli_subject = nats_client.new_inbox()

        await _publish(
            nats_client,
            msg_type=_MessageType.CONNECT_REQUEST,
            body=json_dumpb(self._connect_id),
            subject=self._get_connect_subject_name(),
            reply=hs_to_cli_subject
        )

        resp_waiter = \
            await MessageWaiter.new(nats_client, hs_to_cli_subject, no_ack=True)

        message = await resp_waiter.wait()
        if message.type != _MessageType.CONNECT_RESPONSE:
            raise ValueError('Unknown message on connect response subject')

        if message.reply is None:
            raise ValueError('Server subject not specified in connect response')
        hs_to_srv_subject = message.reply

        resp_waiter.set_timeout_duration(10)

        try:
            conn = await NATSConnection._do_handshake_and_create_connection(
                nats_client,
                self,
                handshaker,
                resp_waiter,
                hs_to_srv_subject
            )
        except Exception:
            if nats_client.is_connected:
                await resp_waiter.unsubscribe()
            raise

        message = await resp_waiter.wait()
        await resp_waiter.unsubscribe()

        if message.type != _MessageType.HS_DONE:
            raise ValueError(f'Got message with "{message.type}" type '
                              'instead of handshake confirmation')

        to_cli_subject, to_srv_subject = \
            self._get_transport_subject_names(self._connect_id)  # type: ignore
        policy = ClientCheckPolicy(conn, self.CONNECTION_CHECK_PERIOD * 1.5)
        conn._set_run_data(policy, to_cli_subject, to_srv_subject)

        return conn

    async def client_connect(self, handshaker: Handshaker) -> BaseConnection:
        if self._connect_id is None:
            raise TypeError('Cannot connect to server. For those, who connect, '
                            'connect_id parameter must be set. '
                            'Check NATSConnector docs for details.')

        client = await self._get_client()
        return await self._client_connect(handshaker, client)

    async def cleanup(self):
        """
        Drop all pending messages and
        delete subjects with connector's connect_id

        If connect_id is None, noop
        """
        return

    def dump_state(self) -> str:
        """Unsupported for this connector"""
        raise TypeError('This method unsupported cause user and password '
                        'are not encrypted in dump, which is insecure')

    @classmethod
    def from_state(cls, state: str) -> Self:
        """Unsupported for this connector"""
        raise TypeError('This method unsupported cause user and password '
                        'are not encrypted in dump, which is insecure')

    async def _ack_message(self, nats_client, message):
        await nats_client.basic_ack(message.delivery_tag)

    async def _get_client(self):
        key = (asyncio.get_running_loop(), self._urls)

        if (client := _clients.get(key)) is None:
            fut = asyncio.get_running_loop().create_future()
            _clients[key] = fut
            try:
                client = await nats.connect(list(self._urls))
            except Exception as e:
                fut.set_exception(e)
                del(_clients[key])
                raise
            else:
                fut.set_result(client)
                _clients[key] = client

        elif isinstance(client, asyncio.Future):
            client = await client

        elif client.is_closed:
            del(_clients[key])
            return await self._get_client()

        return client

    async def _check_connection_use(self):
        return  # закрытие соединения, если оно больше никому не нужно



from communica.entities.route import RouteClient, RouteServer
from communica.entities.simple import SimpleClient, SimpleServer


__all__ = (
    'SimpleClient',
    'SimpleServer',
    'RouteClient',
    'RouteServer',
)



import asyncio

from abc import abstractmethod, ABC
from typing import Protocol, Any

from typing_extensions import Self

from communica.utils import HasLoopMixin, logger
from communica.connectors.base import BaseConnector


class BaseEntity(ABC, HasLoopMixin):
    __slots__ = ('connector', '_closed', '_stop_event')

    connector: BaseConnector

    def __init__(self, connector: BaseConnector) -> None:
        self.connector = connector

    def __repr__(self) -> str:
        return f'{type(self).__name__}()'

    async def run(self):
        """Run until .stop() is called."""
        if hasattr(self, '_stop_event'):
            raise RuntimeError('.run() must be called only once on same object')

        self._closed = asyncio.get_event_loop().create_future()
        self._stop_event = asyncio.Event()
        try:
            async with self:
                print(f'Running {self!r} on {self.connector.repr_address()}')
                await self._stop_event.wait()
        finally:
            del(self._stop_event)
            self._closed.set_result(None)

    def stop(self):
        """If .run() was called, it returns, otherwise this method does nothing"""
        if hasattr(self, '_stop_event'):
            self._stop_event.set()

    async def wait_stop(self):
        """Call .stop() and wait until closed"""
        if hasattr(self, '_stop_event'):
            self.stop()
            await self._closed

    @abstractmethod
    async def init(self) -> Self:
        raise NotImplementedError

    @abstractmethod
    async def close(self) -> None:
        raise NotImplementedError

    async def __aenter__(self) -> Self:
        return await self.init()

    async def __aexit__(self, exc_type, exc, tb) -> None:
        await self.close()


class BaseClient(BaseEntity):
    __slots__ = ()

    @abstractmethod
    async def init(self, timeout: 'int | None' = None) -> Self:
        """
        Establish connection with server.

        On connect timeout cancels connect task.

        Args:
            timeout: Time in seconds,
              if omitted, method will block until connect succeed.

        Raises:
            TimeoutError
        """
        raise NotImplementedError

    @abstractmethod
    async def close(self) -> None:
        """Close connection with server."""
        raise NotImplementedError


class BaseServer(BaseEntity):
    __slots__ = ()

    @abstractmethod
    async def init(self) -> Self:
        """Create server and start accepting connections"""
        raise NotImplementedError

    @abstractmethod
    async def close(self) -> None:
        """Stop serving and cancel all handler tasks"""
        raise NotImplementedError


class SyncHandlerType(Protocol):
    def __call__(self, data: Any) -> Any: ...


class AsyncHandlerType(Protocol):
    async def __call__(self, data: Any) -> Any: ...



import asyncio
from uuid import uuid4
from typing import Any, TypedDict

from communica.utils import TaskSet, logger, fmt_task_name
from communica.exceptions import ReqError
from communica.serializers import BaseSerializer, default_serializer
from communica.entities.base import SyncHandlerType, AsyncHandlerType
from communica.connectors.base import (
    BaseConnector,
)
from communica.entities.simple import (
    RequestType,
    ReqRepClient,
    ReqRepServer,
    RequestHandler,
    ReqRepMessageFlow,
)


__all__ = (
    'RouteClient',
    'RouteServer'
)


class Metadata(TypedDict):
    id: str
    type: RequestType
    route: str


class RouteMessageFlow(ReqRepMessageFlow):
    __slots__ = ('routes', 'fallback_task_set', '_response_serializers',)

    routes: 'dict[str, tuple[RequestHandler, BaseSerializer]]'
    fallback_task_set: TaskSet

    _response_serializers: 'dict[str, BaseSerializer]'

    def __init__(self):
        super().__init__()
        self.routes = {}
        self.fallback_task_set = TaskSet()
        self._response_serializers = {}

    def update_route(
            self,
            route: str,
            handler: 'SyncHandlerType | AsyncHandlerType',
            serializer: 'BaseSerializer | None'
    ):
        if serializer is None:
            serializer = default_serializer
        self.routes[route] = (RequestHandler(handler), serializer)

    def dispatch(self, metadata: Metadata, raw_data: bytes):
        if metadata['type'] < RequestType.RESP_OK:
            try:
                route_handle = self.routes[metadata['route']]
                task_set = route_handle[0].running_tasks
            except KeyError:
                route_handle = None
                task_set = self.fallback_task_set

            task_set.create_task_with_exc_log(
                self.handle_request(route_handle, metadata, raw_data),
                name=fmt_task_name('route-request-handler')
            )
        else:
            try:
                serializer = self._response_serializers[metadata['route']]
            except KeyError:  # pragma: no cover
                # this is possible after program restart.
                # just log and skip, cause we don't have a routine,
                # waiting for this response
                # XXX: should it be logged?
                logger.warning('Got response with not requested '
                              f'route ({metadata["route"]})')
                return
            self._handle_response(serializer, metadata, raw_data)

    async def handle_request(
            self,
            route_handle: 'tuple[RequestHandler, BaseSerializer] | None',
            req_meta: Metadata,
            raw_data: bytes
        ):
        if route_handle is not None:
            handler, serializer = route_handle
            resp_meta, resp_data = await self._handle_request(
                handler, serializer, req_meta, raw_data
            )
        else:
            serializer = default_serializer
            resp_meta = req_meta.copy()
            resp_meta['type'] = RequestType.RESP_ERR_REQUESTER
            resp_data = \
                ReqError(f'Route "{req_meta["route"]}" not exists').to_dict()

        await self._send_response(req_meta, resp_meta, resp_data, serializer)

    async def request(
            self,
            route: str,
            serializer: 'BaseSerializer | None',
            data: Any
    ) -> bytes:
        request_id, fut = self._create_response_waiter()

        serializer = serializer or default_serializer
        if (saved_serializer := self._response_serializers.get(route)) is None:
            self._response_serializers[route] = serializer
        elif saved_serializer is not serializer:
            raise TypeError('You should always use the same serializer for '
                           f'the same route ({saved_serializer!r} '
                            'has already been used earlier)')

        await self._connection.send(
            Metadata(id=request_id, type=RequestType.REQ_REP, route=route),
            serializer.client_dump(data)
        )

        return await fut

    async def throw(
            self,
            route: str,
            serializer: 'BaseSerializer | None',
            data: Any
    ) -> None:
        await self._connection.send(
            Metadata(id='', type=RequestType.REQ_THROW, route=route),
            (serializer or default_serializer).client_dump(data)
        )


class RouteClient(ReqRepClient[RouteMessageFlow]):
    """
    Pair for RouteServer.

    Handler searched by exact match of route string.
    """

    __slots__ = ()

    def __init__(
            self,
            connector: BaseConnector,
            client_id: 'str | None' = None
    ) -> None:
        super().__init__(connector)

        self._flow = RouteMessageFlow()

        self._run_task = None
        self._client_id = client_id or uuid4().hex

    def route_handler(
            self,
            route: str,
            serializer: 'BaseSerializer | None' = None
    ):
        """
        Register handler for route

        Args:
            route: Handler identifier.
            serializer: Instance of :obj:`communica.serializers.BaseSeralizer`.
                Defaults to JsonSerializer.
        """
        def decorator(endpoint: 'SyncHandlerType | AsyncHandlerType'):
            self._flow.update_route(route, endpoint, serializer)

        return decorator

    async def request(
            self,
            route: str,
            data: Any,
            serializer: 'BaseSerializer | None' = None
    ) -> Any:
        """
        Send request, wait response.

        Args:
            route: Handler identifier.
            client_id: If omitted or None, random client will be chosen.
            serializer: Instance of :obj:`communica.serializers.BaseSeralizer`.
                Defaults to JsonSerializer.
        """
        return await self._flow.request(route, serializer, data)

    async def throw(
            self,
            route: str,
            data: Any,
            serializer: 'BaseSerializer | None' = None
    ) -> None:
        """
        Send request without waiting response.

        Args:
            route: Handler identifier.
            client_id: If omitted or None, random client will be chosen.
            serializer: Instance of :obj:`communica.serializers.BaseSeralizer`.
                Defaults to JsonSerializer.
        """
        return await self._flow.throw(route, serializer, data)


class RouteServer(ReqRepServer[RouteMessageFlow]):
    """
    Pair for RouteClient.

    Handler searched by exact match of route string.
    """

    __slots__ = ('_routes',)

    _known_clients: 'dict[str, RouteMessageFlow | asyncio.Future[RouteMessageFlow]]'

    def __init__(
            self,
            connector: BaseConnector,
    ) -> None:
        self.connector = connector
        self._routes = []
        self._known_clients = {}
        self._client_connected = asyncio.Event()
        self._client_conn_runners = {}

    def route_handler(
            self,
            route: str,
            serializer: 'BaseSerializer | None' = None
    ):
        """
        Register handler for route

        Args:
            route: Handler identifier.
            serializer: Instance of :obj:`communica.serializers.BaseSeralizer`.
                Defaults to JsonSerializer.
        """
        def decorator(endpoint: 'SyncHandlerType | AsyncHandlerType'):
            for flow in self._known_clients.values():
                if isinstance(flow, RouteMessageFlow):
                    flow.update_route(route, endpoint, serializer)
            self._routes.append((route, endpoint, serializer))

        return decorator

    def _cancel_handler_tasks(self, flow: RouteMessageFlow):
        flow.fallback_task_set.cancel()

        for handler, _ in flow.routes.values():
            handler.running_tasks.cancel()

    def _create_new_flow(self) -> RouteMessageFlow:
        flow = RouteMessageFlow()
        for route, handler, serializer in self._routes:
            flow.update_route(route, handler, serializer)
        return flow

    async def request(
            self,
            route: str,
            data: Any,
            serializer: 'BaseSerializer | None' = None,
            client_id: 'str | None' = None
    ) -> Any:
        """
        Send request, wait response.

        Args:
            route: Handler identifier.
            client_id: If omitted or None, random connected client will be chosen.
            serializer: Instance of :obj:`communica.serializers.BaseSeralizer`.
                Defaults to JsonSerializer. Once serializer was used for
                request, no other serializer can be used for same route.
        """
        flow = await self._get_client_flow(client_id)
        return await flow.request(route, serializer, data)

    async def throw(
            self,
            route: str,
            data: Any,
            serializer: 'BaseSerializer | None' = None,
            client_id: 'str | None' = None
    ) -> None:
        """
        Send request without waiting response.

        Args:
            route: Handler identifier.
            client_id: If omitted or None, random connected client will be chosen.
            serializer: Instance of :obj:`communica.serializers.BaseSeralizer`.
                Defaults to JsonSerializer. Once serializer was used for
                request, no other serializer can be used for same route.
        """
        flow = await self._get_client_flow(client_id)
        return await flow.throw(route, serializer, data)
